@@ -5,13 +5,12 @@ import json
 import logging
 from abc import abstractmethod
 from dataclasses import dataclass
+from typing import Optional, Union
 from urllib.parse import urlunparse
 
-import pandas as pd
-import pytz
+import polars as pl
 
 from wetterdienst.metadata.columns import Columns
-from wetterdienst.util.pandas import chunker
 from wetterdienst.util.url import ConnectionString
 
 log = logging.getLogger(__name__)
@@ -26,44 +25,52 @@ class ExportMixin:
     acquired through the core machinery.
     """
 
-    df: pd.DataFrame
+    df: pl.DataFrame
 
-    def fill_gaps(self):
-        self.df = self.df.fillna(-999)
-        return self.df
-
-    def filter_by_sql(self, sql: str) -> pd.DataFrame:
+    def filter_by_sql(self, sql: str) -> pl.DataFrame:
         self.df = self._filter_by_sql(self.df, sql)
         return self.df
 
-    def to_dict(self) -> dict:
+    @abstractmethod
+    def to_dict(self, *args, **kwargs) -> dict:
+        pass
 
-        # Convert all datetime columns to ISO format.
-        df = convert_datetimes(self.df)
+    @abstractmethod
+    def to_json(self, *args, **kwargs) -> str:
+        pass
 
-        # Return dictionary with timeseries types.
-        return df.to_dict(orient="records")
+    @abstractmethod
+    def to_ogc_feature_collection(self, *args, **kwargs) -> dict:
+        pass
 
-    def to_json(self, indent: int = 4):
-        df = self.df.copy()
-
-        for column in ("from_date", "to_date", "date"):
-            if column in df:
-                df[column] = df[column].map(lambda x: x.isoformat() if pd.notna(x) and x is not None else None)
-
-        return df.to_json(orient="records", date_format="iso", indent=indent, force_ascii=False)
-
-    def to_csv(self):
-        return self.df.to_csv(index=False, date_format="%Y-%m-%dT%H-%M-%S")
-
-    def to_geojson(self, indent: int = 4) -> str:
+    def to_geojson(self, with_metadata: bool = False, indent: Optional[Union[int, bool]] = 4, **_kwargs) -> str:
         """
-        Convert station information into GeoJSON format.
-
-        Return:
-             JSON string in GeoJSON FeatureCollection format.
+        Convert station information into GeoJSON format
+        :param with_metadata: Include metadata in GeoJSON output
+        :param indent: Indentation for JSON output
+        :param _kwargs: Additional arguments passed to `to_ogc_feature_collection`
+        :return: JSON string in GeoJSON FeatureCollection format
         """
-        return json.dumps(self.to_ogc_feature_collection(), indent=indent, ensure_ascii=False)
+        if indent is True:
+            indent = 4
+        elif indent is False:
+            indent = None
+        json_kwargs = {"indent": indent, "ensure_ascii": False}
+        return json.dumps(self.to_ogc_feature_collection(with_metadata=with_metadata), **json_kwargs)
+
+    def to_csv(self, **kwargs) -> str:
+        """
+        :param kwargs: Additional arguments passed to `pl.DataFrame.write_csv`
+        :return: CSV string
+        """
+        df = self.df
+        date_columns = (Columns.START_DATE.value, Columns.END_DATE.value, Columns.DATE.value)
+        df = df.with_columns(
+            pl.col(column).map_elements(lambda date: date.isoformat() if date else None, return_dtype=pl.Utf8)
+            for column in date_columns
+            if column in df.columns
+        )
+        return df.write_csv(**kwargs)
 
     def to_format(self, fmt: str, **kwargs) -> str:
         """
@@ -75,16 +82,16 @@ class ExportMixin:
         fmt = fmt.lower()
 
         if fmt == "json":
-            return self.to_json(indent=kwargs.get("indent"))
+            return self.to_json(**kwargs)
         elif fmt == "csv":
             return self.to_csv()
         elif fmt == "geojson":
-            return self.to_geojson(indent=kwargs.get("indent"))
+            return self.to_geojson(**kwargs)
         else:
             raise KeyError("Unknown output format")
 
     @staticmethod
-    def _filter_by_sql(df: pd.DataFrame, sql: str) -> pd.DataFrame:
+    def _filter_by_sql(df: pl.DataFrame, sql: str) -> pl.DataFrame:
         """
         Filter Pandas DataFrame using an SQL query.
         The virtual table name is "data", so queries
@@ -100,23 +107,9 @@ class ExportMixin:
         """
         import duckdb
 
-        df = duckdb.query_df(df, "data", sql).df()
-
-        for column in (
-            Columns.FROM_DATE.value,
-            Columns.TO_DATE.value,
-            Columns.DATE.value,
-        ):
-            try:
-                df[column] = df[column].dt.tz_convert(pytz.UTC)
-            except KeyError:
-                pass
-
-        return df
-
-    @abstractmethod
-    def to_ogc_feature_collection(self):
-        pass
+        df = df.with_columns(pl.col(Columns.DATE.value).dt.replace_time_zone(None)).to_pandas()
+        df = duckdb.query_df(df, "data", sql).pl()
+        return df.with_columns(pl.col(Columns.DATE.value).dt.replace_time_zone("UTC"))
 
     def to_target(self, target: str):
         """
@@ -139,29 +132,26 @@ class ExportMixin:
         :return: self
         """
 
-        log.info(f"Exporting records to {target}\n{self.df.count()}")
+        log.info(f"Exporting records to {target}\n{self.df.select(pl.count())}")
 
         connspec = ConnectionString(target)
-        protocol = connspec.url.scheme
-        database = connspec.get_database()
-        tablename = connspec.get_table()
+        protocol = connspec.protocol
+        database = connspec.database
+        tablename = connspec.table
 
         if target.startswith("file://"):
-            filepath = connspec.get_path()
+            filepath = connspec.path
 
             if target.endswith(".xlsx"):
                 log.info(f"Writing to spreadsheet file '{filepath}'")
-
                 # Convert all datetime columns to ISO format.
                 df = convert_datetimes(self.df)
-                df.to_excel(filepath, index=False)
+                df.write_excel(filepath)
 
             elif target.endswith(".feather"):
                 # https://arrow.apache.org/docs/python/feather.html
                 log.info(f"Writing to Feather file '{filepath}'")
-                import pyarrow.feather as feather
-
-                feather.write_feather(self.df, filepath, compression="lz4")
+                self.df.write_ipc(filepath, compression="lz4")
 
             elif target.endswith(".parquet"):
                 """
@@ -178,11 +168,8 @@ class ExportMixin:
                 """  # noqa:E501
 
                 log.info(f"Writing to Parquet file '{filepath}'")
-                import pyarrow as pa
-                import pyarrow.parquet as pq
 
-                table = pa.Table.from_pandas(self.df)
-                pq.write_table(table, filepath)
+                self.df.write_parquet(filepath)
 
             elif target.endswith(".zarr"):
                 """
@@ -198,33 +185,29 @@ class ExportMixin:
                 log.info(f"Writing to Zarr group '{filepath}'")
                 import xarray
 
-                df = self.df
-
-                # Problem: `ValueError: Cannot setitem on a Categorical with a new category, set the categories first`.
-                # Solution: Let's convert all categorical columns back to their designated type representations.
-                #           https://stackoverflow.com/questions/32011359/convert-categorical-data-in-pandas-dataframe/32011969#32011969
-                if "quality" in df:
-                    df.quality = df.quality.astype("Int64")
-                categorical_columns = df.select_dtypes(["category"]).columns
-                df[categorical_columns] = df[categorical_columns].astype("str")
-
                 # Problem: `TypeError: float() argument must be a string or a number, not 'NAType'`.
                 # Solution: Fill gaps in the data.
-                df = df.fillna(-999)
+                df = self.df.fill_null(-999)
+                df = df.with_columns(
+                    pl.col("date")
+                    .dt.convert_time_zone("UTC")
+                    .dt.replace_time_zone(None)
+                    .map_elements(lambda date: date.isoformat())
+                )
+                group = df.get_column("dataset").gather(0).item()
+                df = df.to_pandas()
 
                 # Convert pandas DataFrame to xarray Dataset.
                 dataset = xarray.Dataset.from_dataframe(df)
                 log.info(f"Converted to xarray Dataset. Size={dataset.sizes}")
 
                 # Export to Zarr format.
-                # TODO: Add "group" parameter.
-                #       Group path. (a.k.a. `path` in zarr terminology.)
                 # TODO: Also use attributes: `store.set_attribute()`
                 store = dataset.to_zarr(
                     filepath,
                     mode="w",
-                    group=None,
-                    encoding={"date": {"dtype": "datetime64"}},
+                    group=group,  # use dataset name as group name
+                    encoding={"date": {"dtype": "datetime64[ns]"}},
                 )
 
                 # Reporting.
@@ -262,8 +245,10 @@ class ExportMixin:
             log.info(f"Writing to DuckDB. database={database}, table={tablename}")
             import duckdb
 
+            df = self.df.with_columns(pl.col(Columns.DATE.value).dt.replace_time_zone(None)).to_pandas()
+
             connection = duckdb.connect(database=database, read_only=False)
-            connection.register("origin", self.df)
+            connection.register("origin", df)
             connection.execute(f"DROP TABLE IF EXISTS {tablename};")
             connection.execute(f"CREATE TABLE {tablename} AS SELECT * FROM origin;")  # noqa:S608
 
@@ -320,6 +305,28 @@ class ExportMixin:
                 alias fetch="wetterdienst values --provider=dwd --network=observation --parameter=kl --resolution=daily --period=recent --station=1048,4411"
                 fetch --target="influxdb2://${INFLUXDB_ORGANIZATION}:${INFLUXDB_TOKEN}@localhost/?database=dwd&table=weather"
 
+            Example queries::
+
+                influx query 'from(bucket:"dwd") |> range(start:-2d) |> limit(n: 10)'
+
+
+            ==========================
+            InfluxDB 3.x database sink
+            ==========================
+
+            Install Python driver::
+
+                pip install influxdb3_python
+
+            Acquire data::
+
+                INFLUXDB_ORGANIZATION=acme
+                INFLUXDB_TOKEN=t5PJry6TyepGsG7IY_n0K4VHp5uPvt9iap60qNHIXL4E6mW9dLmowGdNz0BDi6aK_bAbtD76Z7ddfho6luL2LA==
+                INFLUXDB_HOST="eu-central-1-1.aws.cloud2.influxdata.com"
+
+                alias fetch="wetterdienst values --provider=dwd --network=observation --parameter=kl --resolution=daily --period=recent --station=1048,4411"
+                fetch --target="influxdb3://${INFLUXDB_ORGANIZATION}:${INFLUXDB_TOKEN}@${INFLUXDB_HOST}/?database=dwd&table=weather"
+
             Example queries::
 
                 influx query 'from(bucket:"dwd") |> range(start:-2d) |> limit(n: 10)'
@@ -329,79 +336,81 @@ class ExportMixin:
                 version = 1
             elif protocol in ["influxdb2", "influxdb2s"]:
                 version = 2
+            elif protocol in ["influxdb3", "influxdb3s"]:
+                version = 3
             else:
                 raise KeyError(f"Unknown protocol variant '{protocol}' for InfluxDB")
 
             log.info(f"Writing to InfluxDB version {version}. database={database}, table={tablename}")
 
-            # 1. Mungle the data frame.
-            # Use the "date" column as appropriate timestamp index.
-            df = self.df.set_index(pd.DatetimeIndex(self.df["date"]))
-            df = df.drop(["date"], axis=1)
-
-            # Compute designated tag fields from some candidates.
-            tag_columns = []
-            tag_candidates = [
-                Columns.STATION_ID.value,
-                Columns.QUALITY.value,
-                Columns.QUALITY_PREFIX.value,
-                Columns.DATASET.value,
-                Columns.PARAMETER.value,
-            ]
-            for tag_candidate in tag_candidates:
-                tag_candidate = tag_candidate.lower()
-                for column in df.columns:
-                    if column.startswith(tag_candidate):
-                        tag_columns.append(column)
-
-            # Setup the connection.
+            # Set up the connection.
             if version == 1:
                 from influxdb import InfluxDBClient
 
                 client = InfluxDBClient(
-                    host=connspec.url.hostname,
-                    port=connspec.url.port or 8086,
-                    username=connspec.url.username,
-                    password=connspec.url.password,
+                    host=connspec.host,
+                    port=connspec.port or 8086,
+                    username=connspec.username,
+                    password=connspec.password,
                     database=database,
                     ssl=protocol.endswith("s"),
                 )
                 client.create_database(database)
             elif version == 2:
-                from influxdb_client import InfluxDBClient, Point
+                from influxdb_client import InfluxDBClient as InfluxDBClientV2
+                from influxdb_client import Point as PointV2
                 from influxdb_client.client.write_api import SYNCHRONOUS
 
                 ssl = protocol.endswith("s")
                 url = f"http{ssl and 's' or ''}://{connspec.url.hostname}:{connspec.url.port or 8086}"
-                client = InfluxDBClient(url=url, org=connspec.url.username, token=connspec.url.password)
+                client = InfluxDBClientV2(url=url, org=connspec.username, token=connspec.password)
                 write_api = client.write_api(write_options=SYNCHRONOUS)
+            elif version == 3:
+                from influxdb_client_3 import (
+                    InfluxDBClient3 as InfluxDBClientV3,
+                )
+                from influxdb_client_3 import (
+                    Point as PointV3,
+                )
+                from influxdb_client_3 import (
+                    WriteOptions,
+                    write_client_options,
+                )
+                from influxdb_client_3.write_client.client.write_api import WriteType
+
+                write_options = WriteOptions(write_type=WriteType.synchronous)
+                wco = write_client_options(WriteOptions=write_options)
+                client_v3 = InfluxDBClientV3(
+                    host=connspec.host,
+                    org=connspec.username,
+                    token=connspec.password,
+                    write_client_options=wco,
+                    database=database,
+                )
 
             points = []
-            for items in chunker(df, chunksize=50000):
+            for record in self.df.iter_rows(named=True):
+                # for record in items.iter_rows(named=True):
+                time = record.pop("date").isoformat()
+                fields = {k: v for k, v in record.items() if v is not None}
+                if not fields:
+                    continue
+                if version == 1:
+                    point = {
+                        "measurement": tablename,
+                        "time": time,
+                        "fields": fields,
+                    }
+                elif version == 2:
+                    point = PointV2(tablename).time(time)
+                    for field, value in fields.items():
+                        point = point.field(field, value)
+                elif version == 3:
+                    point = PointV3(tablename).time(time)
+                    for field, value in fields.items():
+                        point = point.field(field, value)
 
-                for date, record in items.iterrows():
-                    time = date.isoformat()
-                    tags = {tag: record.pop(tag) for tag in tag_columns if tag in record}
-
-                    fields = record.dropna().to_dict()
-                    if not fields:
-                        continue
-
-                    if version == 1:
-                        point = {
-                            "measurement": tablename,
-                            "time": time,
-                            "tags": tags,
-                            "fields": fields,
-                        }
-                    elif version == 2:
-                        point = Point(tablename).time(date.isoformat())
-                        for tag, value in tags.items():
-                            point = point.tag(tag, value)
-                        for field, value in fields.items():
-                            point = point.field(field, value)
-
-                    points.append(point)
+                points.append(point)
 
             # Write to InfluxDB.
             if version == 1:
@@ -412,6 +421,8 @@ class ExportMixin:
             elif version == 2:
                 write_api.write(bucket=database, record=points)
                 write_api.close()
+            elif version == 3:
+                client_v3.write(record=points)
 
             log.info("Writing to InfluxDB finished")
 
@@ -450,9 +461,9 @@ class ExportMixin:
 
             # Convert timezone-aware datetime fields to naive ones.
             # FIXME: Omit this as soon as the CrateDB driver is capable of supporting timezone-qualified timestamps.
-            self.df.date = self.df.date.dt.tz_localize(None)
+            df = self.df.with_columns(pl.col("date").dt.replace_time_zone(time_zone=None))
 
-            self.df.to_sql(
+            df.to_pandas().to_sql(
                 name=tablename,
                 con=cratedb_target,
                 schema=database,
@@ -475,7 +486,8 @@ class ExportMixin:
             Examples::
 
                 # Prepare
-                alias fetch='wetterdienst dwd observation values --station=1048,4411 --parameter=kl --resolution=daily --period=recent'
+                alias fetch='wetterdienst dwd observation values --station=1048,4411 --parameter=kl --resolution=daily
+                --period=recent'
 
                 # Acquire data.
                 fetch --target="sqlite:///dwd.sqlite?table=weather"
@@ -483,44 +495,42 @@ class ExportMixin:
                 # Query data.
                 sqlite3 dwd.sqlite "SELECT * FROM weather;"
 
-            """  # noqa:E501
-
+            """
             # Honour SQLite's SQLITE_MAX_VARIABLE_NUMBER, which defaults to 999
             # for SQLite versions prior to 3.32.0 (2020-05-22),
             # see https://www.sqlite.org/limits.html#max_variable_number.
-            chunksize = 5000
+            chunk_size = 5000
             if target.startswith("sqlite://"):
                 import sqlite3
 
                 if sqlite3.sqlite_version_info < (3, 32, 0):
-                    chunksize = int(999 / len(self.df.columns))
+                    chunk_size = int(999 / len(self.df.columns))
 
             log.info("Writing to SQL database")
-            self.df.to_sql(
+            self.df.to_pandas().to_sql(
                 name=tablename,
                 con=target,
                 if_exists="replace",
                 index=False,
                 method="multi",
-                chunksize=chunksize,
+                chunksize=chunk_size,
             )
             log.info("Writing to SQL database finished")
 
 
-def convert_datetimes(df: pd.DataFrame) -> pd.DataFrame:
+def convert_datetimes(df: pl.DataFrame) -> pl.DataFrame:
     """
     Convert all datetime columns to ISO format.
 
-    :param df:        df[Columns.FROM_DATE] = df[Columns.FROM_DATE].dt.tz_localize(self.tz)
+    :param df:        df[Columns.START_DATE] = df[Columns.START_DATE].dt.tz_localize(self.tz)
     :return:
     """
-    df: pd.DataFrame = df.copy(deep=True)
-
-    date_columns = list(df.select_dtypes(include=[pd.DatetimeTZDtype]).columns)
-    date_columns.extend([Columns.FROM_DATE.value, Columns.TO_DATE.value, Columns.DATE.value])
+    date_columns = list(df.select(pl.col(pl.Datetime)).columns)
+    date_columns.extend([Columns.START_DATE.value, Columns.END_DATE.value, Columns.DATE.value])
     date_columns = set(date_columns)
     for date_column in date_columns:
         if date_column in df:
-            df[date_column] = df[date_column].apply(lambda d: d.isoformat() if pd.notna(d) else None)
-
+            df = df.with_columns(
+                pl.col(date_column).map_elements(lambda v: v.isoformat() if v else None, return_dtype=pl.Utf8)
+            )
     return df
