@@ -4,6 +4,7 @@
 from typing import Tuple, Union
 
 import numpy as np
+import polars as pl
 from sklearn.neighbors import BallTree
 
 
@@ -64,16 +65,33 @@ def derive_nearest_neighbours(
         Tuple of distances and ranks of nearest to most distant stations_result
     """
     points = np.c_[np.radians(latitudes), np.radians(longitudes)]
-
     distance_tree = BallTree(points, metric="haversine")
-
     return distance_tree.query(coordinates.get_coordinates_in_radians().reshape(-1, 2), k=number_nearby)
 
 
-def convert_dm_to_dd(dms: float) -> float:
-    """Convert degree minutes to decimal degree"""
-    degrees, minutes = divmod(dms, 1)
-
-    decimals = round(minutes / 60 * 100, 2)
-
+def convert_dm_to_dd(dm: pl.Series) -> pl.Series:
+    """
+    Convert degree minutes (floats) to decimal degree
+    :param dm:
+    :return:
+    """
+    degrees = dm.cast(int)
+    minutes = dm - degrees
+    decimals = (minutes / 60 * 100).round(2)
     return degrees + decimals
+
+
+def convert_dms_string_to_dd(dms: pl.Series) -> pl.Series:
+    """
+    Convert degree minutes seconds (string) to decimal degree
+    e.g. 49 18 21
+    :param dms:
+    :return:
+    """
+    dms = dms.str.split(" ").to_frame("dms")
+    dms = dms.select(
+        pl.col("dms").list.get(0).cast(pl.Float64).alias("degrees"),
+        pl.col("dms").list.get(1).cast(pl.Float64).alias("minutes"),
+        pl.col("dms").list.get(2).cast(pl.Float64).alias("seconds"),
+    )
+    return dms.get_column("degrees").rename("") + (dms.get_column("minutes") / 60) + (dms.get_column("seconds") / 3600)
